@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2010 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -476,7 +476,7 @@ mktime_utc (struct tm *tm)
 	tm->tm_isdst = -1;
 	tt = mktime (tm);
 	
-#if defined (G_OS_WIN32)
+#if defined (G_OS_WIN32) && !defined (__MINGW32__)
 	_get_timezone (&tz);
 	if (tm->tm_isdst > 0) {
 		int dst;
@@ -1541,7 +1541,11 @@ charset_convert (iconv_t cd, const char *inbuf, size_t inleft, char **outp, size
 				outbuf = out + rc;
 			}
 			
-			if (errno == EINVAL || errno == EILSEQ) {
+			/* Note: GnuWin32's libiconv 1.9 can also set errno to ERANGE
+			 * which seems to mean that it encountered a character that
+			 * does not fit the specified 'from' charset. We'll handle
+			 * that the same way we handle EILSEQ. */
+			if (errno == EILSEQ || errno == ERANGE) {
 				/* invalid or incomplete multibyte
 				 * sequence in the input buffer */
 				*outbuf++ = '?';
@@ -1566,6 +1570,7 @@ charset_convert (iconv_t cd, const char *inbuf, size_t inleft, char **outp, size
 
 #define USER_CHARSETS_INCLUDE_UTF8    (1 << 0)
 #define USER_CHARSETS_INCLUDE_LOCALE  (1 << 1)
+#define USER_CHARSETS_INCLUDE_LATIN1  (1 << 2)
 
 
 /**
@@ -1594,30 +1599,38 @@ g_mime_utils_decode_8bit (const char *text, size_t len)
 	g_return_val_if_fail (text != NULL, NULL);
 	
 	locale = g_mime_locale_charset ();
-	if (locale && !g_ascii_strcasecmp (locale, "UTF-8"))
+	if (!g_ascii_strcasecmp (locale, "iso-8859-1") ||
+	    !g_ascii_strcasecmp (locale, "UTF-8")) {
+		/* If the user's locale charset is either of these, we
+		 * don't need to include the locale charset in our list
+		 * of fallback charsets. */
 		included |= USER_CHARSETS_INCLUDE_LOCALE;
+	}
 	
 	if ((user_charsets = g_mime_user_charsets ())) {
 		while (user_charsets[i])
 			i++;
 	}
 	
-	charsets = g_alloca (sizeof (char *) * (i + 3));
+	charsets = g_alloca (sizeof (char *) * (i + 4));
 	i = 0;
 	
 	if (user_charsets) {
 		while (user_charsets[i]) {
 			/* keep a record of whether or not the user-supplied
-			 * charsets include UTF-8 and/or the default fallback
+			 * charsets include UTF-8, Latin1, or the user's locale
 			 * charset so that we avoid doubling our efforts for
-			 * these 2 charsets. We could have used a hash table
+			 * these 3 charsets. We could have used a hash table
 			 * to keep track of unique charsets, but we can
 			 * (hopefully) assume that user_charsets is a unique
 			 * list of charsets with no duplicates. */
+			if (!g_ascii_strcasecmp (user_charsets[i], "iso-8859-1"))
+				included |= USER_CHARSETS_INCLUDE_LATIN1;
+			
 			if (!g_ascii_strcasecmp (user_charsets[i], "UTF-8"))
 				included |= USER_CHARSETS_INCLUDE_UTF8;
 			
-			if (locale && !g_ascii_strcasecmp (user_charsets[i], locale))
+			if (!g_ascii_strcasecmp (user_charsets[i], locale))
 				included |= USER_CHARSETS_INCLUDE_LOCALE;
 			
 			charsets[i] = user_charsets[i];
@@ -1631,6 +1644,9 @@ g_mime_utils_decode_8bit (const char *text, size_t len)
 	if (!(included & USER_CHARSETS_INCLUDE_LOCALE))
 		charsets[i++] = locale;
 	
+	if (!(included & USER_CHARSETS_INCLUDE_LATIN1))
+		charsets[i++] = "iso-8859-1";
+	
 	charsets[i] = NULL;
 	
 	min = len;
@@ -1802,10 +1818,10 @@ rfc2047_decode_word (const char *in, size_t inlen)
 		p = (char *) decoded;
 		len = declen;
 		
-		//while (!g_utf8_validate (p, len, (const char **) &p)) {
-		//	len = declen - (p - (char *) decoded);
-		//	*p = '?';
-		//}
+		while (!g_utf8_validate (p, len, (const char **) &p)) {
+			len = declen - (p - (char *) decoded);
+			*p = '?';
+		}
 		
 		return g_strndup ((char *) decoded, declen);
 	}
@@ -1877,7 +1893,7 @@ g_mime_utils_header_decode_text (const char *text)
 			word = inptr;
 			ascii = TRUE;
 			
-			if (enable_rfc2047_workarounds) {
+			if (G_UNLIKELY (enable_rfc2047_workarounds)) {
 				if (!strncmp (inptr, "=?", 2)) {
 					inptr += 2;
 					
@@ -1899,8 +1915,15 @@ g_mime_utils_header_decode_text (const char *text)
 						inptr++;
 					}
 					
-					if (!strncmp (inptr, "?=", 2))
-						inptr += 2;
+					if (*inptr == '\0') {
+						/* didn't find an end marker... */
+						inptr = word + 2;
+						ascii = TRUE;
+						
+						goto non_rfc2047;
+					}
+					
+					inptr += 2;
 				} else {
 				non_rfc2047:
 					/* stop if we encounter a possible rfc2047 encoded
@@ -1981,9 +2004,10 @@ g_mime_utils_header_decode_text (const char *text)
 char *
 g_mime_utils_header_decode_phrase (const char *phrase)
 {
+	gboolean enable_rfc2047_workarounds = _g_mime_enable_rfc2047_workarounds ();
 	register const char *inptr = phrase;
 	gboolean encoded = FALSE;
-	const char *lwsp, *text;
+	const char *lwsp, *word;
 	size_t nlwsp, n;
 	gboolean ascii;
 	char *decoded;
@@ -2001,14 +2025,53 @@ g_mime_utils_header_decode_phrase (const char *phrase)
 		
 		nlwsp = (size_t) (inptr - lwsp);
 		
-		text = inptr;
+		word = inptr;
 		if (is_atom (*inptr)) {
-			while (is_atom (*inptr))
-				inptr++;
+			if (G_UNLIKELY (enable_rfc2047_workarounds)) {
+				/* Make an extra effort to detect and
+				 * separate encoded-word tokens that
+				 * have been merged with other
+				 * words. */
+				
+				if (!strncmp (inptr, "=?", 2)) {
+					inptr += 2;
+					
+					/* skip past the charset (if one is even declared, sigh) */
+					while (*inptr && *inptr != '?')
+						inptr++;
+					
+					/* sanity check encoding type */
+					if (inptr[0] != '?' || !strchr ("BbQq", inptr[1]) || inptr[2] != '?')
+						goto non_rfc2047;
+					
+					inptr += 3;
+					
+					/* find the end of the rfc2047 encoded word token */
+					while (*inptr && strncmp (inptr, "?=", 2) != 0)
+						inptr++;
+					
+					if (*inptr == '\0') {
+						/* didn't find an end marker... */
+						inptr = word + 2;
+						goto non_rfc2047;
+					}
+					
+					inptr += 2;
+				} else {
+				non_rfc2047:
+					/* stop if we encounter a possible rfc2047 encoded
+					 * token even if it's inside another word, sigh. */
+					while (is_atom (*inptr) && strncmp (inptr, "=?", 2) != 0)
+						inptr++;
+				}
+			} else {
+				while (is_atom (*inptr))
+					inptr++;
+			}
 			
-			n = (size_t) (inptr - text);
-			if (is_rfc2047_encoded_word (text, n)) {
-				if ((decoded = rfc2047_decode_word (text, n))) {
+			n = (size_t) (inptr - word);
+			if (is_rfc2047_encoded_word (word, n)) {
+				if ((decoded = rfc2047_decode_word (word, n))) {
 					/* rfc2047 states that you must ignore all
 					 * whitespace between encoded words */
 					if (!encoded)
@@ -2032,20 +2095,20 @@ g_mime_utils_header_decode_phrase (const char *phrase)
 			g_string_append_len (out, lwsp, nlwsp);
 			
 			ascii = TRUE;
-			while (*inptr && !is_lwsp (*inptr)) {
+			while (*inptr && !is_lwsp (*inptr) && !is_atom (*inptr)) {
 				ascii = ascii && is_ascii (*inptr);
 				inptr++;
 			}
 			
-			n = (size_t) (inptr - text);
+			n = (size_t) (inptr - word);
 			
 			if (!ascii) {
 				/* *sigh* I hate broken mailers... */
-				decoded = g_mime_utils_decode_8bit (text, n);
+				decoded = g_mime_utils_decode_8bit (word, n);
 				g_string_append (out, decoded);
 				g_free (decoded);
 			} else {
-				g_string_append_len (out, text, n);
+				g_string_append_len (out, word, n);
 			}
 			
 			encoded = FALSE;
